import math
import re

import actions
import predefinedentities
from jstypes import JSArray, JSObject, JSWrapper

from validator.constants import BUGZILLA_BUG
from validator.compat import FX48_DEFINITION

# Function prototypes should implement the following:
#  wrapper : The JSWrapper instace that is being called
#  arguments : A list of argument nodes; untraversed
#  traverser : The current traverser object


def webbrowserpersist(wrapper, arguments, traverser):
    """
    Most nsIWebBrowserPersist should no longer be used, in favor of the new
    Downloads.jsm interfaces.
    """
    traverser.err.warning(
        err_id=('testcases_javascript_call_definititions',
                'webbrowserpersist'),
        warning='nsIWebBrowserPersist should no longer be used',
        description=('Most nsIWebBrowserPersist methods have been '
                     'superseded by simpler methods in Downloads.jsm, namely '
                     '`Downloads.fetch` and `Downloads.createDownload`. See '
                     'http://mzl.la/downloads-jsm for more information.'),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context,
        tier=4)


def webbrowserpersist_saveuri(wrapper, arguments, traverser):
    """
    nsIWebBrowserPersist.saveURI requires a valid privacy context as
    of Firefox 19
    """
    if len(arguments) >= 7:
        load_context = traverser._traverse_node(arguments[6])
        if load_context.get_literal_value() is None:
            traverser.err.warning(
                err_id=('testcases_javascript_call_definititions',
                        'webbrowserpersist_saveuri'),
                warning=('saveURI should not be called with a null load '
                         'context'),
                description=('While nsIWebBrowserPersist.saveURI accepts null '
                             'in place of a privacy context, this usage is '
                             'acceptable only when no appropriate load '
                             'context exists.'),
                filename=traverser.filename,
                line=traverser.line,
                column=traverser.position,
                context=traverser.context,
                tier=4)

    webbrowserpersist(wrapper, arguments, traverser)


def xpcom_constructor(method, extend=False, mutate=False, pretraversed=False):
    """Returns a function which wraps an XPCOM class instantiation function."""

    def definition(wrapper, arguments, traverser):
        """Wraps an XPCOM class instantiation function."""

        if not arguments:
            return None

        traverser._debug('(XPCOM Encountered)')

        if not pretraversed:
            arguments = [traverser._traverse_node(x) for x in arguments]
        argz = arguments[0]

        if not argz.is_global or 'xpcom_map' not in argz.value:
            argz = JSWrapper(traverser=traverser)
            argz.value = {'xpcom_map': lambda: {'value': {}}}

        traverser._debug('(Building XPCOM...)')

        inst = traverser._build_global(
            method, argz.value['xpcom_map']())
        inst.value['overwritable'] = True

        if extend or mutate:
            # FIXME: There should be a way to get this without
            # traversing the call chain twice.
            parent = actions.trace_member(traverser, wrapper['callee']['object'])

            if mutate and not (parent.is_global and
                               isinstance(parent.value, dict) and
                               'value' in parent.value):
                # Assume that the parent object is a first class
                # wrapped native
                parent.value = inst.value

                # FIXME: Only objects marked as global are processed
                # as XPCOM instances
                parent.is_global = True

            if isinstance(parent.value, dict):
                if extend and mutate:
                    if callable(parent.value['value']):
                        parent.value['value'] = \
                            parent.value['value'](t=traverser)

                    parent.value['value'].update(inst.value['value'])
                    return parent

                if extend:
                    inst.value['value'].update(parent.value['value'])

                if mutate:
                    parent.value = inst.value

        return inst
    definition.__name__ = 'xpcom_%s' % str(method)
    return definition


# Global object function definitions:
def string_global(wrapper, arguments, traverser):
    if not arguments:
        return JSWrapper('', traverser=traverser)
    arg = traverser._traverse_node(arguments[0])
    value = actions._get_as_str(arg.get_literal_value())
    return JSWrapper(value, traverser=traverser)


def array_global(wrapper, arguments, traverser):
    output = JSArray()
    if arguments:
        output.elements = [traverser._traverse_node(a) for a in arguments]
    return JSWrapper(output, traverser=traverser)


def number_global(wrapper, arguments, traverser):
    if not arguments:
        return JSWrapper(0, traverser=traverser)
    arg = traverser._traverse_node(arguments[0])
    try:
        value = float(arg.get_literal_value())
    except (ValueError, TypeError):
        return traverser._build_global(
                name='NaN',
                entity=predefinedentities.GLOBAL_ENTITIES[u'NaN'])
    return JSWrapper(value, traverser=traverser)


def boolean_global(wrapper, arguments, traverser):
    if not arguments:
        return JSWrapper(False, traverser=traverser)
    arg = traverser._traverse_node(arguments[0])
    return JSWrapper(bool(arg.get_literal_value()), traverser=traverser)


def python_wrap(func, args, nargs=False):
    """
    This is a helper function that wraps Python functions and exposes them to
    the JS engine. The first parameter should be the Python function to wrap.
    The second parameter should be a list of tuples. Each tuple should
    contain:

     1. The type of value to expect:
        - "string"
        - "num"
     2. A default value.
    """

    def _process_literal(type_, literal):
        if type_ == 'string':
            return actions._get_as_str(literal)
        elif type_ == 'num':
            return actions._get_as_num(literal)
        return literal

    def wrap(wrapper, arguments, traverser):
        passed_args = [traverser._traverse_node(a) for a in arguments]

        params = []
        if not nargs:
            # Handle definite argument lists.
            for type_, def_value in args:
                if passed_args:
                    parg = passed_args[0]
                    passed_args = passed_args[1:]

                    passed_literal = parg.get_literal_value()
                    passed_literal = _process_literal(type_, passed_literal)
                    params.append(passed_literal)
                else:
                    params.append(def_value)
        else:
            # Handle dynamic argument lists.
            for arg in passed_args:
                literal = arg.get_literal_value()
                params.append(_process_literal(args[0], literal))

        traverser._debug('Calling wrapped Python function with: (%s)' %
                             ', '.join(map(str, params)))
        try:
            output = func(*params)
        except (ValueError, TypeError, OverflowError):
            # If we cannot compute output, just return nothing.
            output = None

        return JSWrapper(output, traverser=traverser)

    return wrap


def math_log(wrapper, arguments, traverser):
    """Return a better value than the standard python log function."""
    args = [traverser._traverse_node(a) for a in arguments]
    if not args:
        return JSWrapper(0, traverser=traverser)

    arg = actions._get_as_num(args[0].get_literal_value())
    if arg == 0:
        return JSWrapper(float('-inf'), traverser=traverser)

    if arg < 0:
        return JSWrapper(traverser=traverser)

    arg = math.log(arg)
    return JSWrapper(arg, traverser=traverser)


def math_random(wrapper, arguments, traverser):
    """Return a "random" value for Math.random()."""
    return JSWrapper(0.5, traverser=traverser)


def math_round(wrapper, arguments, traverser):
    """Return a better value than the standard python round function."""
    args = [traverser._traverse_node(a) for a in arguments]
    if not args:
        return JSWrapper(0, traverser=traverser)

    arg = actions._get_as_num(args[0].get_literal_value())
    # Prevent nasty infinity tracebacks.
    if abs(arg) == float('inf'):
        return args[0]

    # Python rounds away from zero, JS rounds "up".
    if arg < 0 and int(arg) != arg:
        arg += 0.0000000000000001
    arg = round(arg)
    return JSWrapper(arg, traverser=traverser)


def nsIJSON_deprec(wrapper, arguments, traverser):
    """Throw a compatibility error about removed XPCOM methods."""
    traverser.notice(
        err_id=('testcases_javascript_calldefinitions', 'nsIJSON',
                'deprec'),
        notice='Deprecated nsIJSON methods in use.',
        description=('The `encode` and `decode` methods in nsIJSON have been '
                     'deprecated in Gecko 7. You can use the methods in the '
                     'global JSON object instead. See %s for more '
                     'information.') %
                         'https://developer.mozilla.org/En/Using_native_JSON')

    return JSWrapper(JSObject(), traverser=traverser, dirty=True)


def js_wrap(wrapper, arguments, traverser):
    """Return the wrapped variant of an unwrapped JSObject."""
    if not arguments:
        traverser._debug('WRAP:NO ARGS')
        return

    traverser._debug('WRAPPING OBJECT')
    obj = traverser._traverse_node(arguments[0])
    if obj.value is None:
        traverser._debug('WRAPPING OBJECT>>NOTHING TO WRAP')
        return JSWrapper(JSObject(), traverser=traverser)

    if len(arguments) > 1:
        traverser.warning(
            err_id=('testcases_js_xpcom', 'xpcnativewrapper', 'shallow'),
            warning='Shallow XPCOM wrappers should not be used',
            description='Shallow XPCOM wrappers are seldom necessary and '
                        'should not be used. Please use deep wrappers '
                        'instead.',
            signing_help='Extensions making use of shallow wrappers will not '
                         'be accepted for automated signing. Please remove '
                         'the second and subsequent arguments of any calls '
                         'to `XPCNativeWrapper`, as well as any code which '
                         'applies `XPCNativeWrapper` to properties obtained '
                         'from these shallowly wrapped objects.',
            signing_severity='high')
        # Do not mark shallow wrappers as not unwrapped.
        return obj

    if obj.is_global:
        # Why are we changing the original object? XPCNativeWrapper
        # does not alter its arguments.
        obj.value['is_unwrapped'] = False
    else:
        obj.value.is_unwrapped = False

    return obj


def js_unwrap(wrapper, arguments, traverser):
    """Return the unwrapped variant of an unwrapped JSObject."""
    if not arguments:
        traverser._debug('UNWRAP:NO ARGS')
        return

    traverser._debug('UNWRAPPING OBJECT')
    obj = traverser._traverse_node(arguments[0])
    if obj.value is None:
        traverser._debug('UNWRAPPING OBJECT>>NOTHING TO UNWRAP')
        return JSWrapper(JSObject(unwrapped=True), traverser=traverser)

    if obj.is_global:
        obj.value['is_unwrapped'] = True
    else:
        obj.value.is_unwrapped = True

    return obj


def open_in_chrome_context(uri, method, traverser):
    if not uri.is_literal():
        traverser.err.notice(
            err_id=('js', 'instanceactions', '%s_nonliteral' % method),
            notice='`%s` called with non-literal parameter.' % method,
            description='Calling `%s` with variable parameters can result in '
                        'potential security vulnerabilities if the variable '
                        'contains a remote URI. Consider using `window.open` '
                        'with the `chrome=no` flag.' % method,
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)

    remote_url = re.compile(r'^(https?|ftp|data):(//)?', re.I)
    uri = unicode(uri.get_literal_value())
    if uri.startswith('//') or remote_url.match(uri):
        traverser.err.warning(
            err_id=('js', 'instanceactions', '%s_remote_uri' % method),
            warning='`%s` called with non-local URI.' % method,
            description='Calling `%s` with a non-local URI will result in the '
                        'dialog being opened with chrome privileges.' % method,
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)


def Proxy_deprec(wrapper, arguments, traverser):
    traverser.warning(
        err_id=('testcases_javascript_calldefinitions', 'Proxy', 'deprec'),
        warning='Proxy.create and Proxy.createFunction are no longer supported.',
        description=(
            'Proxy.create and Proxy.createFunction are no longer supported. '
            'If this flag appears on Add-ons SDK code, make sure you download '
            'the latest version of the SDK and submit a new version. '
            'See %s for more information.' % BUGZILLA_BUG % 892903),
        compatibility_type='error',
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context,
        for_appversions=FX48_DEFINITION)



from helper import CompatTestCase
from validator.compat import FX53_DEFINITION


class TestFX53Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 53 are properly executed."""

    VERSION = FX53_DEFINITION

    def test_splitmenu_element_removed(self):
        """https://github.com/mozilla/amo-validator/issues/506"""
        expected = 'The splitmenu element has been removed.'

        invalid = '''
            splitmenu {
                -moz-binding: url("chrome://browser/content/urlbarBindings.xml#splitmenu");
            }
        '''

        valid = '''
            .splitmenu-menuitem {
              -moz-binding: url("chrome://global/content/bindings/menu.xml#menuitem");
              list-style-image: inherit;
              -moz-image-region: inherit;
            }
        '''

        self.run_regex_for_compat(invalid, is_js=False)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

        self.reset()

        self.run_regex_for_compat(valid, is_js=False)

        assert not self.compat_err.notices
        assert not self.compat_err.errors
        assert not self.compat_err.warnings

    def test_moz_calc_removed(self):
        """https://github.com/mozilla/amo-validator/issues/505"""
        expected = 'The -moz-calc function has been removed.'

        invalid = [
            'min-width: -moz-calc(6em - 14px);',
            'background-size: -moz-calc(100% - 2px) -moz-calc(100% - 2px);',
            '-moz-radial-gradient(-moz-calc(100px + -25px) top, red, blue)',
        ]

        for script in invalid:
            self.run_regex_for_compat(script, is_js=False)
            assert not self.compat_err.notices
            assert not self.compat_err.errors
            self.assert_compat_warning('warning', expected)

            self.reset()

    def test_geturiforkeyword_removed(self):
        """https://github.com/mozilla/amo-validator/issues/504"""
        expected = 'The getURIForKeyword function was removed.'

        script = '''
          let bmserv = Cc["@mozilla.org/browser/nav-bookmarks-service;1"].
                       getService(Ci.nsINavBookmarksService);
          if (bmserv.getURIForKeyword(alias.value))
            bduplicate = true;
        '''

        self.run_script_for_compat(script)
        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

    def test_openuriinnewntab_changed(self):
        """https://github.com/mozilla/amo-validator/issues/503"""
        expected = (
            'The _openURIInNewTab function was changed and now requires '
            'an nsIURI for the referrer.')

        script = '''
          let _original__openURIInNewTab = nsBrowserAccess._openURIInNewTab;

          nsBrowserAccess.prototype._openURIInNewTab = function(aURI, aReferrer, aReferrerPolicy, aIsPrivate,
                                                                aIsExternal, aForceNotRemote=false) {
            let browser = _original__openURIInNewTab.apply(nsBrowserAccess.prototype, arguments);
            if (!aIsExternal) {
              browser.setAttribute("diverted", "true");
            }
            return browser;
          };
        '''

        self.run_script_for_compat(script)
        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

    def test_nsIX509CertDB(self):
        """https://github.com/mozilla/amo-validator/issues/502"""
        expected = (
            'The nsIX509CertDB interface was changed so it no longer '
            'exposes the certificate nickname.')

        script = '''
            XmlMultiDSigUtil.mX509CertDB = Components.classes["@mozilla.org/security/x509certdb;1"]
                            .getService(Components.interfaces.nsIX509CertDB);
            signerCert = XmlMultiDSigUtil.mX509CertDB.findCertByNickname(null, signerCertNickname);
        '''

        self.run_script_for_compat(script)
        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

    def test_nsIPK11TokenDBfindTokenByName(self):
        """https://github.com/mozilla/amo-validator/issues/507"""
        expected = 'Calling findTokenByName("") is no longer valid.'

        script = '''
            Components.classes["@mozilla.org/security/pk11tokendb;1"]
                .getService(Components.interfaces.nsIPK11TokenDB)
                .findTokenByName("").logoutAndDropAuthenticatedResources();
        '''

        self.run_script_for_compat(script)
        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

        # Make sure that we explicitly check for an empty first argument
        script = '''
            Components.classes["@mozilla.org/security/pk11tokendb;1"]
                .getService(Components.interfaces.nsIPK11TokenDB)
                .findTokenByName("not empty")
                .logoutAndDropAuthenticatedResources();
        '''

        self.run_script_for_compat(script)
        assert not self.compat_err.notices
        assert not self.compat_err.errors
        assert not self.compat_err.warnings



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Derived from nsVersionComparator.cpp in mozilla-central

from itertools import izip_longest
import re


__all__ = 'Version', 'VersionPart'


def strcmp(a, b):
    # Null string comes after any non-null string
    if not a:
        return 1 if b else 0
    elif not b:
        return -1

    return cmp(a, b)


class VersionPart(object):
    numA   = 0
    strB   = ''
    numC   = 0
    extraD = ''

    def __init__(self, part):
        self._part = part

        if part == '*':
            self.numA = float('inf')
        else:
            self.numA, self.strB = self._splitnum(part)

        if self.strB:
            if self.strB[0] == '+':
                self.numA += 1
                self.strB = 'pre'
            else:
                match = re.match(r'([^\d+-]*)(.*)', self.strB)

                self.strB = match.group(1)
                self.numC, self.extraD = self._splitnum(match.group(2))

    def __repr__(self):
        return 'VersionPart(%s)' % repr(self._part)

    def __str__(self):
        return self._part

    def __cmp__(self, other):
        if other is None:
            return 1

        return (cmp(self.numA, other.numA) or
                strcmp(self.strB, other.strB) or
                cmp(self.numC, other.numC) or
                strcmp(self.extraD, other.extraD))

    def _splitnum(self, string):
        match = re.match(r'((?:[+-]?\d+)?)(.*)', string)
        return (int(match.group(1) or 0),
                match.group(2))


class Version(object):

    def __init__(self, version):
        self._version = version
        self.parts = map(VersionPart, version.split('.'))

    def __repr__(self):
        return 'Version(%s)' % repr(self._version)

    def __str__(self):
        return self._version

    def __cmp__(self, other):
        for s_part, o_part in izip_longest(self.parts, other.parts):
            cres = cmp(s_part, o_part)
            if cres:
                return cres
        return 0

    @property
    def is_release(self):
        return bool(re.match(r'^[\d.]+$', self._version))



import re
import cssutils

from validator.constants import PACKAGE_THEME
from validator.contextgenerator import ContextGenerator
from validator.unicodehelper import NON_ASCII_FILTER


BAD_URL_PAT = ("url\(['\"]?(?!(chrome:|resource:))(\/\/|(ht|f)tps?:\/\/|data:)"
               ".*['\"]?\)")
BAD_URL = re.compile(BAD_URL_PAT, re.I)
REM_URL = re.compile("url\(['\"]?(\/\/|ht|f)tps?:\/\/.*['\"]?\)", re.I)

SKIP_TYPES = ('S', 'COMMENT')

DOWNLOADS_INDICATOR_BUG = 845408


def test_css_file(err, filename, data, line_start=1):
    'Parse and test a whole CSS file.'

    tokenizer = cssutils.tokenize2.Tokenizer()
    context = ContextGenerator(data)

    if data:
        # Remove any characters which aren't printable, 7-bit ASCII.
        data = NON_ASCII_FILTER.sub('', data)

    token_generator = tokenizer.tokenize(data)

    _run_css_tests(err,
                   tokens=token_generator,
                   filename=filename,
                   line_start=line_start - 1,
                   context=context)


def test_css_snippet(err, filename, data, line):
    'Parse and test a CSS nugget.'

    # Re-package to make it CSS-complete. Note the whitespace to prevent
    # the extra code from showing in the context output.
    data = '#foo{\n\n%s\n\n}' % data

    test_css_file(err, filename, data, line)


def _run_css_tests(err, tokens, filename, line_start=0, context=None):
    """Processes a CSS file to test it for things that could cause it
    to be harmful to the browser."""

    last_descriptor = None

    identity_box_mods = []
    unicode_errors = []
    downloads_indicator_selectors = []

    token_history = []

    while True:

        try:
            tok_type, value, line, position = tokens.next()
        except UnicodeDecodeError:
            unicode_errors.append(str(line + line_start))
            continue
        except StopIteration:
            break

        if tok_type in SKIP_TYPES:
            continue

        token_history.append((tok_type, value))

        # Save the last descriptor for reference.
        if tok_type in ('IDENT', 'FUNCTION'):
            if tok_type == 'IDENT':
                last_descriptor = value.lower()

        elif tok_type == 'URI':

            # If we hit a URI after -moz-binding, we may have a
            # potential security issue.
            if last_descriptor == '-moz-binding' and BAD_URL.match(value):
                # We need to make sure the URI is not remote.
                err.warning(
                    err_id=('css', '_run_css_tests', '-moz-binding_external'),
                    warning='Illegal reference to external scripts',
                    description='`-moz-binding` may not reference external '
                                'scripts in CSS. This is considered to be a '
                                'security issue. The script file must be '
                                'placed in the /content/ directory of the '
                                'package.',
                    filename=filename,
                    line=line + line_start,
                    context=context.get_context(line))
            elif (err.detected_type == PACKAGE_THEME and
                  REM_URL.match(value) and
                  token_history[-2][0] != 'NAMESPACE_SYM'):
                # If we're a theme and the URL is remote and the last token
                # wasn't a CSS namespace symbol, raise a warning.
                err.warning(
                    err_id=('css', '_run_css_tests', 'remote_url'),
                    warning='Themes may not reference remote resources',
                    description='Themes may not reference resources that are '
                                'stored on remote servers. All resources must '
                                'be stored locally within the theme.',
                    filename=filename,
                    line=line + line_start,
                    context=context.get_context(line))

        elif tok_type == 'HASH':
            # Search for interference with the identity box.
            if value == '#identity-box':
                identity_box_mods.append(str(line + line_start))
            elif value == '#downloads-indicator':
                downloads_indicator_selectors.append(str(line + line_start))

    if identity_box_mods and err.detected_type != PACKAGE_THEME:
        err.warning(('testcases_markup_csstester', '_run_css_tests',
                     'identity_box'),
                    'Modification to identity box.',
                    ['The identity box (#identity-box) is a sensitive piece '
                     'of the interface and should not be modified.',
                     'Lines: %s' % ', '.join(identity_box_mods)],
                    filename)

    if unicode_errors:
        err.info(('testcases_markup_csstester',
                  'test_css_file',
                  'unicode_decode'),
                 'Unicode decode error.',
                 ['While decoding a CSS file, an unknown character was '
                  'encountered, causing some problems.',
                  'Lines: %s' % ', '.join(unicode_errors)],
                 filename)



import validator.constants
from validator import decorator
from validator.constants import FF4_MIN, APPLICATIONS

APP_VERSIONS_URL = 'Please check the list of valid versions at: '\
        'https://addons.mozilla.org/en-US/firefox/pages/appversions/'


@decorator.register_test(tier=1)
def test_targetedapplications(err, xpi_package=None):
    """Tests to make sure that the targeted applications in the
    manifest file are legit and that any associated files (I'm
    looking at you, SeaMonkey) are where they need to be."""

    manifest_json = err.get_resource('manifest_json')
    install_rdf = err.get_resource('install_rdf')

    if install_rdf:
        applications = install_rdf.get_applications()
        manifest_file = 'install.rdf'
    elif manifest_json:
        applications = manifest_json.get_applications()
        manifest_file = 'manifest.json'
    else:  # No manifest.json, no install.rdf file.
        if err.supported_versions is None:
            err.supported_versions = {}
        return

    APPROVED_APPLICATIONS = validator.constants.APPROVED_APPLICATIONS
    # Compute APPROVED_IDS dynamically here because APPROVED_APPLICATIONS
    # is a changing constant (updated on the fly on validate.validate).
    APPROVED_IDS = dict((x['guid'], y) for (y, x)
                        in APPROVED_APPLICATIONS.items()
                        if x['guid'] in APPLICATIONS)

    used_targets = []
    all_supported_versions = {}

    err.metadata['applications'] = {}

    # Isolate all of the bnodes referring to target applications
    for target_app in applications:
        guid = target_app['guid']

        used_targets.append(guid)

        found_id = APPROVED_IDS.get(guid)

        if found_id:
            # Remember if the addon supports Firefox.
            is_firefox = APPLICATIONS[guid] == 'firefox'

            # Grab the minimum and maximum version numbers.
            if (err.overrides and
                    'targetapp_minVersion' in err.overrides and
                    guid in err.overrides['targetapp_minVersion']):
                min_version = err.overrides['targetapp_minVersion'][guid]
            else:
                min_version = target_app.get('min_version')

            if (err.overrides and
                    'targetapp_maxVersion' in err.overrides and
                    guid in err.overrides['targetapp_maxVersion']):
                max_version = err.overrides['targetapp_maxVersion'][guid]
            else:
                max_version = target_app.get('max_version')

            # Get the approved app versions for this application.
            app_versions = APPROVED_APPLICATIONS[found_id]['versions']

            # Ensure that the version numbers are in the app's
            # list of acceptable version numbers.

            app_name = APPLICATIONS[guid] if guid in APPLICATIONS else guid

            if min_version is None:
                err.error(('testcases_targetapplication',
                           'test_targetedapplications',
                           'missing_minversion'),
                          'Missing minVersion property',
                          'A target application element is missing its '
                          'min version property.',
                          filename=manifest_file)
                continue
            elif max_version is None:
                err.error(('testcases_targetapplication',
                           'test_targetedapplications',
                           'missing_maxversion'),
                          'Missing maxVersion property',
                          'A target application element is missing its '
                          'max version property.',
                          filename=manifest_file)
                continue

            try:
                min_ver_pos = app_versions.index(min_version)
            except ValueError:
                err.error(('testcases_targetapplication',
                           'test_targetedapplications',
                           'invalid_min_version'),
                          'Invalid minimum version number',
                          ['The minimum version that was specified is not '
                           'an acceptable version number for the Mozilla '
                           'product that it corresponds with.',
                           'Version "%s" isn\'t compatible with "%s".' %
                              (min_version, app_name),
                           APP_VERSIONS_URL],
                          manifest_file)
                continue

            try:
                max_ver_pos = app_versions.index(max_version)
            except ValueError:
                err.error(('testcases_targetapplication',
                           'test_targetedapplications',
                           'invalid_max_version'),
                          'Invalid maximum version number',
                          ['The maximum version that was specified is not '
                           'an acceptable version number for the Mozilla '
                           'product that it corresponds with.',
                           'Version "%s" isn\'t compatible with "%s".' %
                              (max_version, app_name),
                           APP_VERSIONS_URL],
                          manifest_file)
                continue

            # Now we need to check to see if the version numbers
            # are in the right order.
            if min_ver_pos > max_ver_pos:
                err.error(('testcases_targetapplication',
                           'test_targetedapplications',
                           'invalid_version_order'),
                          'Invalid min/max versions',
                          ['The version numbers provided for the '
                           'application in question are not in the '
                           'correct order. The maximum version must be '
                           'greater than the minimum version.',
                           '"%s" is not less than "%s".' % (min_version,
                                                            max_version)],
                          manifest_file)
                continue

            all_supported_versions[guid] = (
                app_versions[min_ver_pos:max_ver_pos + 1])

            err.metadata['applications'][APPLICATIONS[guid]] = {
                'min': min_version,
                'max': max_version,
            }

            # Test whether it's a FF4 addon

            # NOTE: This should probably also be extrapolated for
            # Thunderbird and the like when they get up to speed. The tests
            # will likely be the same down the line, so we can keep the
            # "ff4" resource as a legacy thing and worry about it later.
            if is_firefox:
                ff4_pos = app_versions.index(FF4_MIN)
                if max_ver_pos >= ff4_pos:
                    err.save_resource('ff4', True)

    no_duplicate_targets = set(used_targets)

    if len(used_targets) != len(no_duplicate_targets):
        err.error(('testcases_targetapplication',
                   'test_targetedapplications',
                   'duplicate_targetapps'),
                  'Found duplicate target application elements.',
                  'Multiple target application elements were found in the '
                  'manifest file that refer to the same application GUID. '
                  'There should not be duplicate target applications '
                  'entries.',
                  'install.rdf')

    # This finds the UUID of the supported applications and puts it in
    # a fun and easy-to-use format for use in other tests.
    final_guids = map(str, no_duplicate_targets)
    approved_guids = list(set(final_guids).intersection(APPROVED_IDS.keys()))
    supports = [APPLICATIONS[k] for k in approved_guids]
    err.save_resource('supports', supports)

    if not err.supported_versions:  # if not set for compatibility
        err.supported_versions = all_supported_versions

    if not supports and err.get_resource('listed'):
        err.error(
            err_id=('testcases_targetapplication',
                    'test_targetedapplications',
                    'no_mozilla_support'),
            error='No Mozilla products listed as target applications',
            description=(
                'None of the target applications listed in the manifest are '
                'supported Mozilla products. At least one official Mozilla '
                'product must be supported for inclusion on '
                'addons.mozilla.org.',
                'See https://addons.mozilla.org/firefox/pages/appversions/'
                ' for more information on supported target applications on '
                'AMO.'),
            filename=manifest_file)



import pytest

from helper import CompatTestCase
from validator.compat import FX50_DEFINITION


class TestFX50Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 50 are properly executed."""

    VERSION = FX50_DEFINITION

    def test_nsIX509Cert_removed(self):
        """https://github.com/mozilla/addons-server/issues/3408"""
        expected = (
            'The methods getUsagesArray, requestUsagesArrayAsync, and '
            'getUsagesString are no longer supported.')

        script = '''
            var certdb = Components.classes["@mozilla.org/security/x509certdb;1"].
                getService(Components.interfaces.nsIX509CertDB);
            var cert = certdb.findCertByDBKey(
                tgt.parentNode.firstChild.getAttribute("display"),
                null);

            cert.QueryInterface(Components.interfaces.nsIX509Cert);
            cert.getUsagesArray();
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_nsIX509Cert_requestUsagesArrayAsync_removed(self):
        """https://github.com/mozilla/addons-server/issues/3408"""
        expected = (
            'The methods getUsagesArray, requestUsagesArrayAsync, and '
            'getUsagesString are no longer supported.')

        script = '''
            var certdb = Components.classes["@mozilla.org/security/x509certdb;1"].
                getService(Components.interfaces.nsIX509CertDB);
            var cert = certdb.findCertByDBKey(
                tgt.parentNode.firstChild.getAttribute("display"),
                null);

            cert.QueryInterface(Components.interfaces.nsIX509Cert2);
            cert.requestUsagesArrayAsync();
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_nsIX509Cert_getUsagesString(self):
        """https://github.com/mozilla/addons-server/issues/3408"""
        expected = (
            'The methods getUsagesArray, requestUsagesArrayAsync, and '
            'getUsagesString are no longer supported.')

        script = '''
            var certdb = Components.classes["@mozilla.org/security/x509certdb;1"].
                getService(Components.interfaces.nsIX509CertDB);
            var cert = certdb.findCertByDBKey(
                tgt.parentNode.firstChild.getAttribute("display"),
                null);

            cert.QueryInterface(Components.interfaces.nsIX509Cert3);
            cert.getUsagesString();
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_events_removed_html_xul(self):
        """https://github.com/mozilla/addons-server/issues/3407"""
        expected = (
            'The draggesture and dragdrop events are no longer supported. '
            'You should use the standard equivalents instead: '
            'dragstart and drop.')

        events = ('draggesture', 'dragdrop', 'ondraggesture', 'ondragdrop')

        base = '''
            <treechildren
                {0}="nsDragAndDrop.startDrag(event, foxclocks.mainManager)"/>
        '''

        for event in events:
            self.run_regex_for_compat(base.format(event), is_js=False)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            warning = [
                msg for msg in self.compat_err.warnings
                if msg['message'].startswith(expected)]
            assert warning
            assert warning[0]['compatibility_type'] == 'error'

    def test_events_removed_js(self):
        """https://github.com/mozilla/addons-server/issues/3407"""
        expected = (
            'The draggesture and dragdrop events are no longer supported. '
            'You should use the standard equivalents instead: '
            'dragstart and drop.')

        events = ('draggesture', 'dragdrop', 'ondraggesture', 'ondragdrop')

        base = '''
            $.on('{0}');
        '''

        for event in events:
            self.run_regex_for_compat(base.format(event))

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            warning = [
                msg for msg in self.compat_err.warnings
                if msg['message'].startswith(expected)]
            assert warning
            assert warning[0]['compatibility_type'] == 'error'



import pytest

from helper import CompatTestCase
from validator.compat import FX51_DEFINITION


class TestFX51Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 50 are properly executed."""

    VERSION = FX51_DEFINITION

    def test_mozIAsyncFavicons_setAndFetchFaviconForPage_changed(self):
        """https://github.com/mozilla/amo-validator/issues/475"""
        expected = (
            'The methods setAndFetchFaviconForPage and '
            'replaceFaviconDataFromDataURL now default to using a null '
            'principal for security reasons. An appropriate principal should '
            'be passed if different behavior is required.')

        script = '''
            var ioService = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService);
            var pageUri = ioService.newURI('http://mozilla.org', 'utf-8', 'mozilla.org');
            var faviconUri = ioService.newURI('http://mozilla.org/fav.ico', 'utf-8', 'mozilla.org');

            var faviconService = Components.classes["@mozilla.org/browser/favicon-service;1"].
                getService(Components.interfaces.nsIFaviconService).
                QueryInterface(Components.interfaces.mozIAsyncFavicons);
            faviconService.setAndFetchFaviconForPage(pageUri, faviconUri, false, faviconService.FAVICON_LOAD_NON_PRIVATE);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'warning'

    def test_mozIAsyncFavicons_replaceFaviconDataFromDataURL_changed(self):
        """https://github.com/mozilla/amo-validator/issues/475"""
        expected = (
            'The methods setAndFetchFaviconForPage and '
            'replaceFaviconDataFromDataURL now default to using a null '
            'principal for security reasons. An appropriate principal should '
            'be passed if different behavior is required.')

        script = '''
            var ioService = Components.classes["@mozilla.org/network/io-service;1"].
                  getService(Components.interfaces.nsIIOService);
            var faviconUri = ioService.newURI('http://mozilla.org/fav.ico', 'utf-8', 'mozilla.org');

            var faviconService = Components.classes["@mozilla.org/browser/favicon-service;1"].
                     getService(Components.interfaces.nsIFaviconService).
                     QueryInterface(Components.interfaces.mozIAsyncFavicons);
            faviconService.replaceFaviconDataFromDataURL(faviconUri, 'data:base64/png');
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'warning'

    def test_BrowserOpenNewTabOrWindow_removed(self):
        """https://github.com/mozilla/amo-validator/issues/473"""
        expected = (
            'The function BrowserOpenNewTabOrWindow has been removed.')

        script = '''
            window.BrowserOpenNewTabOrWindow();
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [
            msg for msg in self.compat_err.warnings
            if msg['message'].startswith(expected)]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_mozVisibilityState_unprefixed(self):
        """https://github.com/mozilla/amo-validator/issues/474"""
        expected = (
            'The mozVisibilityState and mozHidden properties are no longer '
            'prefixed.')

        script = '''
            var x = document.createElement('div');
            x.mozVisibilityState = true;
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [
            msg for msg in self.compat_err.warnings
            if msg['message'].startswith(expected)]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_mozHidden_unprefixed(self):
        """https://github.com/mozilla/amo-validator/issues/474"""
        expected = (
            'The mozVisibilityState and mozHidden properties are no longer '
            'prefixed.')

        script = '''
            var x = document.createElement('div');
            x.mozHidden = true;
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [
            msg for msg in self.compat_err.warnings
            if msg['message'].startswith(expected)]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_onButtonClick_changed(self):
        """https://github.com/mozilla/amo-validator/issues/476"""
        expected = 'The function onButtonClick is now asynchronous.'

        script = '''
            var f = function() { return true };
            var x = document.createElement('div');
            x.onButtonClick(f);
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [
            msg for msg in self.compat_err.warnings
            if msg['message'].startswith(expected)]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'



from ..helper import RegexTestCase
from ..js_helper import TestCase as JSTestCase


class CompatTestCase(JSTestCase, RegexTestCase):
    """
    A TestCase object which aids in running tests regarding compatibility
    tests. It is expected that you will supply a static member `VERSION`
    describing the version(s) of Gecko that these tests apply to.
    """

    def setUp(self, *args, **kwargs):
        super(CompatTestCase, self).setUp(*args, **kwargs)

    def run_script_for_compat(self, script, expose_pollution=False):
        """
        Test a script with and without version restrictions to determine
        whether it properly raises JavaScript compatibility messages.
        """
        self._run_member_for_compat(
            lambda: self.run_script(script, expose_pollution))

    def run_xpcom_for_compat(self, interface, methods):
        """Yields after each method has been run as a script. Used to test that
        XPCOM members are properly flagged for compatibility tests.

        - `interface` should be the name of the XPCOM interface.
        - `methods` should be the member to test. It may be a simple reference
          to a member (i.e.: `foo`) or an action upon a member (i.e.:
          `foo('bar')` or `foo["bar"] = "zap"`).

        """

        script = """
        var x = Components.classes[""].createInstance(Components.interfaces.%s);
        x.%%s;
        """ % interface

        for method in methods:
            self.run_script_for_compat(script % method)
            yield

    def run_regex_for_compat(self, input, is_js=False):
        """Test an input with and without version restrictions to determine
        whether it properly raises regex compatibility messages.

        """
        self._run_member_for_compat(
            lambda: self.run_regex(input, is_js=is_js))

    def _run_member_for_compat(self, method):
        # Run the method without version restrictions.
        self.setup_err()
        method()

        # Store away that error bundle to prepare for the next error bundle.
        standard_err = self.err
        self.err = None

        # Run the method again with version restrictions.
        self.setup_err(for_appversions=self.VERSION)
        method()

        self.compat_err = self.err
        self.err = standard_err

    def assert_compat_silent(self):
        """Assert that no compatibility messages have been raised."""
        assert not any(self.compat_err.compat_summary.values()), \
                'Got %s' % self.compat_err.compat_summary

    def assert_compat_error(self, type_='warning', expected_message=None):
        """Assert that a compat error was raised as a message of type `type_`.

        """
        self._assert_compat_type('error', type_, expected_message)

    def assert_compat_warning(self, type_='notice', expected_message=None):
        """Assert that a compat warning was raised as a message of type
        `type_`.

        """
        self._assert_compat_type('warning', type_, expected_message)

    def _assert_compat_type(self, compat_type, type_, expected_message=None):
        print self.compat_err.print_summary()
        message_collection = {'error': self.compat_err.errors,
                              'warning': self.compat_err.warnings,
                              'notice': self.compat_err.notices}[type_]
        assert message_collection, \
            'No %ss were raised in the compatibility test.' % type_
        is_error = any(
            m['compatibility_type'] == compat_type
            for m in message_collection)

        assert is_error, ('No %ss that raise a compatibility %s were found.' %
                          (type_, compat_type))

        if expected_message is not None:
            has_message = any(
                msg for msg in message_collection
                if msg['message'] == expected_message)
            assert has_message, ('No message \'%s\' found.' % expected_message)



"""
Presumably, the purpose of this module is to load things in an order such
that they do not cause import loops. Given how easy it is, at the moment,
to cause import loops by making trivial changes, this presumption is likely
valid.

It probably also has the effect of making sure each of these modules registers
its testcases in the appropriate places before we begin running tests.
"""
import validator.testcases.chromemanifest
# Import this before content to avoid import loops.
import validator.testcases.javascript.traverser
import validator.testcases.content
import validator.testcases.installrdf
import validator.testcases.jetpack
import validator.testcases.l10ncompleteness
import validator.testcases.langpack
import validator.testcases.packagelayout
import validator.testcases.targetapplication
import validator.testcases.themes



import types

from defusedxml.sax import parse
from rdflib import Graph, URIRef
from rdflib.exceptions import ParserError
from StringIO import StringIO
from xml.sax import ContentHandler, SAXParseException


from validator.constants import APPLICATIONS, APPROVED_APPLICATIONS


class RDFException(Exception):
    """Exception thrown when the RDF parser encounters a problem."""

    def __init__(self, message=None, orig_exception=None):
        if message is None and orig_exception is not None:
            message = orig_exception.getMessage()

        super(RDFException, self).__init__(message)
        self.orig_exception = orig_exception

    def line(self):
        return (self.orig_exception.getLineNumber() if self.orig_exception else
                None)


class AddonRDFEntity(object):
    """
    A "resolved" entity within an RDF file in an add-on. For use by SAX during
    the entity resolution process.
    """

    def getByteStream(self):
        yield None

    def getSystemId(self):
        return ''


class AddonRDFEntityResolver(object):
    """
    An entity resolver to be used by SAX for resolving internal entity
    references.
    """

    def __init__(self, err):
        self.err = err

    def resolveEntity(self, public, system):
        if system.startswith('data:'):
            self.err.warning(
                    err_id=('rdf', 'entity_resolver', 'data_uri'),
                    warning='`data:` URIs are not permitted in `install.rdf`.',
                    filename='install.rdf')
        elif system.startswith('chrome://'):
            self.err.warning(
                    err_id=('rdf', 'entity_resolver', 'chrome_uri'),
                    warning='`chrome://` URI referenced before initialization.',
                    description='A chrome URI was referenced before the '
                                'browser chrome was initialized.',
                    filename='install.rdf')
        else:
            self.err.warning(
                    err_id=('rdf', 'entity_resolver', 'remote_uri'),
                    warning='Remote URI referenced from `install.rdf`.',
                    description='Remote URIs should not be used within '
                                '`install.rdf` files.',
                    filename='install.rdf')

        return AddonRDFEntity()


class RDFParser(object):
    """Parser wrapper for RDF files."""

    def __init__(self, err, data, namespace=None):
        self.err = err
        self.manifest = u'urn:mozilla:install-manifest'
        self.namespace = namespace or 'http://www.mozilla.org/2004/em-rdf'

        if isinstance(data, types.StringTypes):
            data = StringIO(data)  # Wrap data in a pseudo-file

        try:
            # Use an empty ContentHandler, we just want to make sure it parses.
            parse(data, ContentHandler())
        except SAXParseException as ex:
            # Raise the SAX parse exceptions so we get some line info.
            raise RDFException(orig_exception=ex)
        else:
            # We consumed the file, start over from the beginning.
            data.seek(0)

        from rdflib.plugins.parsers import rdfxml
        orig_create_parser = rdfxml.create_parser

        try:
            # Patch rdflib to not resolve URL entities.
            def create_parser(*args, **kwargs):
                parser = orig_create_parser(*args, **kwargs)
                parser.setEntityResolver(AddonRDFEntityResolver(err))
                return parser
            rdfxml.create_parser = create_parser

            # Load up and parse the file in XML format.
            graph = Graph()
            graph.parse(data, format='xml')
            self.rdf = graph

        except ParserError as ex:
            # Re-raise the exception in a local exception type.
            raise RDFException(message=ex.message)
        except SAXParseException as ex:
            # Raise the SAX parse exceptions so we get some line info.
            raise RDFException(orig_exception=ex)
        finally:
            # If we fail, we don't want to sully up the creation function.
            rdfxml.create_parser = orig_create_parser

    def uri(self, element, namespace=None):
        'Returns a URIRef object for use with the RDF document.'

        if namespace is None:
            namespace = self.namespace

        return URIRef('%s#%s' % (namespace, element))

    def get_root_subject(self):
        'Returns the BNode which describes the topmost subject of the graph.'

        manifest = URIRef(self.manifest)

        if list(self.rdf.triples((manifest, None, None))):
            return manifest
        else:
            return self.rdf.subjects(None, self.manifest).next()

    def get_object(self, subject=None, predicate=None):
        """Eliminates some of the glue code for searching RDF. Pass
        in a URIRef object (generated by the `uri` function above or
        a BNode object (returned by this function) for either of the
        parameters."""

        # Get the result of the search
        results = self.rdf.objects(subject, predicate)
        as_list = list(results)

        # Don't raise exceptions, value test!
        if not as_list:
            return None

        return as_list[0]

    def get_objects(self, subject=None, predicate=None):
        """Same as get_object, except returns a list of objects which
        satisfy the query rather than a single result."""

        # Get the result of the search
        results = self.rdf.objects(subject, predicate)
        return list(results)

    def get_applications(self):
        """Return the list of supported applications."""
        applications = []

        # Isolate all of the bnodes referring to target applications
        for target_app in self.get_objects(None,
                                           self.uri('targetApplication')):
            applications.append({
                'guid': self.get_object(target_app, self.uri('id')),
                'min_version': self.get_object(target_app,
                                               self.uri('minVersion')),
                'max_version': self.get_object(target_app,
                                               self.uri('maxVersion'))})
        return applications



from StringIO import StringIO
import unicodehelper


INFINITY = float('inf')


class ContextGenerator:
    """The context generator creates a line-by-line mapping of all files that
    are validated. It will then use that to help produce useful bits of code
    for errors, warnings, and the like."""

    def __init__(self, data=None):
        if isinstance(data, StringIO):
            data = data.getvalue()

        self.data = data.split('\n')

    def get_context(self, line=1, column=0):
        'Returns a tuple containing the context for a line'

        line -= 1  # The line is one-based

        # If there is no data in the file, there can be no context.
        datalen = len(self.data)
        if datalen <= line:
            return None

        build = [self.data[line]]

        # Add surrounding lines if they're available. There must always be
        # three elements in the context.
        if line > 0:
            build.insert(0, self.data[line - 1])
        else:
            build.insert(0, None)

        if line < datalen - 1:
            build.append(self.data[line + 1])
        else:
            build.append(None)

        leading_counts = []

        # Count whitespace to determine how much needs to be stripped.
        lstrip_count = INFINITY
        for line in build:
            # Don't count empty/whitespace-only lines.
            if line is None or not line.strip():
                leading_counts.append(lstrip_count)
                continue

            # Isolate the leading whitespace.
            ws_count = len(line) - len(line.lstrip())
            leading_counts.append(ws_count)
            if ws_count < lstrip_count:
                lstrip_count = ws_count

        # If all of the lines were skipped over, it means everything was
        # whitespace.
        if lstrip_count == INFINITY:
            return ('', '', '')

        for lnum in range(3):
            # Skip edge lines.
            if not build[lnum]:
                continue

            line = build[lnum].strip()

            # Empty lines stay empty.
            if not line:
                build[lnum] = ''
                continue

            line = self._format_line(line, column=column, rel_line=lnum)
            line = '%s%s' % (' ' * (leading_counts[lnum] - lstrip_count), line)

            build[lnum] = line

        # Return the final output as a tuple.
        return tuple(build)

    def _format_line(self, data, column=0, rel_line=1):
        'Formats a line from the data to be the appropriate length'
        line_length = len(data)

        if line_length > 140:
            if rel_line == 0:
                # Trim from the beginning
                data = '... %s' % data[-140:]
            elif rel_line == 1:
                # Trim surrounding the error position
                if column < 70:
                    data = '%s ...' % data[:140]
                elif column > line_length - 70:
                    data = '... %s' % data[-140:]
                else:
                    data = '... %s ...' % data[column - 70:column + 70]

            elif rel_line == 2:
                # Trim from the end
                data = '%s ...' % data[:140]

        data = unicodehelper.decode(data)
        return data

    def get_line(self, position):
        'Returns the line number that the given string position is found on'

        datalen = len(self.data)
        count = len(self.data[0])
        line = 1
        while count < position:
            if line >= datalen:
                break
            count += len(self.data[line]) + 1
            line += 1

        return line




from collections import defaultdict
import itertools
import re
import sys
import types

from validator.constants import DESCRIPTION_TYPES

from . import actions
from .jstypes import JSContext, JSLiteral, JSObject, JSWrapper
from .nodedefinitions import DEFINITIONS
from .predefinedentities import GLOBAL_ENTITIES


DEBUG = False
IGNORE_POLLUTION = False
POLLUTION_COMPONENTS_PATH = re.compile(r'/?components/.*\.jsm?')
POLLUTION_EXCEPTIONS = set(['Cc', 'Ci', 'Cu', ])


class Traverser(object):
    """Traverses the AST Tree and determines problems with a chunk of JS."""

    def __init__(self, err, filename, start_line=0, context=None,
                 is_jsm=False):
        self.err = err
        self.is_jsm = is_jsm

        self.contexts = []
        self.block_contexts = []
        self.filename = filename
        self.start_line = start_line
        self.polluted = False
        self.line = 1  # Line number
        self.position = 0  # Column number
        self.context = context

        # Can use the `this` object
        self.can_use_this = False
        self.this_stack = []

        # For ordering of function traversal.
        self.function_collection = []

        # For debugging
        self.debug_level = 0

        class DebugLevel(object):
            def __enter__(self_):
                self.debug_level += 1

            def __exit__(self_, type, value, traceback):
                self.debug_level -= 1
        self._debug_level = DebugLevel()

    def _debug(self, data, indent=0):
        """Write a message to the console if debugging is enabled."""
        if DEBUG:
            output = data
            if isinstance(data, JSObject) or isinstance(data, JSContext):
                output = data.output()

            output = unicode(output)
            print ('. ' * (self.debug_level + indent) +
                   output.encode('ascii', 'replace'))

        return self._debug_level

    def run(self, data):
        if DEBUG:
            x = open('/tmp/output.js', 'w')
            x.write(unicode(data))
            x.close()

        self._debug('START>>')
        try:
            self.function_collection.append([])
            self._traverse_node(data)

            func_coll = self.function_collection.pop()
            for func in func_coll:
                func()
        except Exception:
            self.system_error(exc_info=sys.exc_info())
            return

        self._debug('END>>')

        if self.contexts:
            # If we're in debug mode, save a copy of the global context for
            # analysis during unit tests.
            if DEBUG:
                self.err.final_context = self.contexts[0]

            if self.pollutable:
                # Ignore anything in the components/ directory
                if POLLUTION_COMPONENTS_PATH.match(self.filename):
                    return

                # This performs the namespace pollution test.
                global_context_size = sum(
                    1 for name in self.contexts[0].data if
                    name not in POLLUTION_EXCEPTIONS)
                self._debug('Final context size: %d' % global_context_size)

                if (global_context_size > 3 and not self.is_jsm and
                        'is_jetpack' not in self.err.metadata and
                        self.err.get_resource('em:bootstrap') != 'true'):
                    self.err.warning(
                        err_id=('testcases_javascript_traverser', 'run',
                                'namespace_pollution'),
                        warning='JavaScript namespace pollution',
                        description=(
                            'Your add-on contains a large number of global '
                            'variables, which may conflict with other '
                            'add-ons. For more information, see '
                            'http://blog.mozilla.com/addons/2009/01/16/'
                            'firefox-extensions-global-namespace-pollution/'
                            ', or use JavaScript modules.',
                            'List of entities: %s'
                            % ', '.join(self.contexts[0].data.keys())),
                        filename=self.filename)

    def _traverse_node(self, node, source=None):
        if node is None:
            return JSWrapper(JSObject(), traverser=self, dirty=True)

        # Simple caching to prevent retraversal
        if '__traversal' in node and node['__traversal'] is not None:
            return node['__traversal']

        if isinstance(node, types.StringTypes):
            return JSWrapper(JSLiteral(node), traverser=self)

        # Extract location information if it's available
        if 'loc' in node and node['loc'] is not None:
            self.line = self.start_line + int(node['loc']['start']['line'])
            self.position = int(node['loc']['start']['column'])

        if node.get('type') not in DEFINITIONS:
            if node.get('type'):
                key = 'unknown_node_types'
                if key not in self.err.metadata:
                    self.err.metadata[key] = defaultdict(int)

                self.err.metadata[key][node['type']] += 1

            return JSWrapper(JSObject(), traverser=self, dirty=True)

        self._debug('TRAVERSE>>%s' % node['type'])
        self.debug_level += 1

        # Extract properties about the node that we're traversing
        (branches, establish_context, action, returns,
         block_level) = DEFINITIONS[node['type']]

        # If we're supposed to establish a context, do it now
        if establish_context:
            self._push_context()
        elif block_level:
            self._push_block_context()

        # An action allows the traverser to make intelligent decisions based
        # on the function of the code, rather than just the content. If an
        # action is availble, run it and store the output.
        action_result = None
        if action is not None:
            action_result = action(self, node)
            # Special case, for immediate literals, define a source property.
            # Used for determining when literals are passed directly
            # as arguments.
            if node['type'] == 'Literal':
                action_result.value.source = source

            if DEBUG:
                action_debug = 'continue'
                if action_result is not None:
                    action_debug = (action_result.output() if
                                    isinstance(action_result, JSWrapper) else
                                    action_result)
                self._debug('ACTION>>%s (%s)' % (action_debug, node['type']))

        if action_result is None:
            self.debug_level += 1
            # Use the node definition to determine and subsequently traverse
            # each of the branches.
            for branch in branches:
                if branch in node:
                    self._debug('BRANCH>>%s' % branch)
                    b = node[branch]
                    if isinstance(b, list):
                        map(self._traverse_node, b)
                    else:
                        self._traverse_node(b)
            self.debug_level -= 1

        # If we defined a context, pop it.
        if establish_context or block_level:
            self._pop_context()
            # WithStatements declare two blocks: one for the block and one for
            # the object that's being withed. We need both because of `let`s.
            if node['type'] == 'WithStatement':
                self._pop_context()

        self.debug_level -= 1

        # If there is an action and the action returned a value, it should be
        # returned to the node traversal that initiated this node's traversal.
        if returns:
            if not isinstance(action_result, JSWrapper):
                return JSWrapper(action_result, traverser=self)
            node['__traversal'] = action_result
            return action_result

        node['__traversal'] = None
        return JSWrapper(JSObject(), traverser=self, dirty=True)

    def _push_block_context(self):
        'Adds a block context to the current interpretation frame'
        self.contexts.append(JSContext('block'))

    def _push_context(self, default=None):
        'Adds a variable context to the current interpretation frame'

        if default is None:
            default = JSContext('default')
        self.contexts.append(default)

        self.debug_level += 1
        self._debug('CONTEXT>>%d' % len(self.contexts))

    def _pop_context(self):
        'Adds a variable context to the current interpretation frame'

        # Keep the global scope on the stack.
        if len(self.contexts) == 1:
            self._debug('CONTEXT>>ROOT POP ABORTED')
            return
        popped_context = self.contexts.pop()

        self.debug_level -= 1
        self._debug('POP_CONTEXT>>%d' % len(self.contexts))
        self._debug(popped_context)

    def _peek_context(self, depth=1):
        """Returns the most recent context. Note that this should NOT be used
        for variable lookups."""

        return self.contexts[len(self.contexts) - depth]

    def _seek_variable(self, variable):
        'Returns the value of a variable that has been declared in a context'

        self._debug('SEEK>>%s' % variable)

        # Look for the variable in the local contexts first
        local_variable = self._seek_local_variable(variable)
        if local_variable is not None:
            # If this variable is also global maintain the xpcom_wildcard value
            # so that it stays linked to its entities. This handles a bug where
            # something like `var Cc = Components.classes` would cause
            # references to `Cc['@some/class']` to not find the right entity.
            if (variable in GLOBAL_ENTITIES and
                    isinstance(local_variable.value, dict)):
                global_variable = self._build_global(
                    variable, GLOBAL_ENTITIES[variable])
                if 'xpcom_wildcard' in global_variable.value:
                    local_variable.value['xpcom_wildcard'] = (
                        global_variable.value['xpcom_wildcard'])
            return local_variable

        # Seek in globals for the variable instead.
        self._debug('SEEK_GLOBAL>>%s' % variable)
        if self._is_global(variable):
            self._debug('SEEK_GLOBAL>>FOUND>>%s' % variable)
            return self._build_global(variable, GLOBAL_ENTITIES[variable])

        self._debug('SEEK_GLOBAL>>FAILED')
        # If we can't find a variable, we always return a dummy object.
        return JSWrapper(JSObject(), traverser=self)

    def _is_defined(self, variable):
        return variable in GLOBAL_ENTITIES or self._is_local_variable(variable)

    def _is_local_variable(self, variable):
        """Return whether a variable is defined in the current scope."""
        return any(ctx.has_var(variable) for ctx in self.contexts)

    def _seek_local_variable(self, variable):
        # Loop through each context in reverse order looking for the defined
        # variable.
        for context in reversed(self.contexts):
            # If it has the variable, return it.
            if context.has_var(variable):
                self._debug('SEEK>>FOUND')
                return context.get(variable, traverser=self)

    def _is_global(self, name):
        'Returns whether a name is a global entity'
        return not self._is_local_variable(name) and name in GLOBAL_ENTITIES

    def _build_global(self, name, entity):
        'Builds an object based on an entity from the predefined entity list'

        if (not callable(entity.get('dangerous')) or
                'dangerous_on_read' in entity):
            dang = entity.get('dangerous', entity.get('dangerous_on_read'))

            if callable(dang):
                dang = dang(self._traverse_node, self.err)

            if dang:
                kwargs = dict(
                    err_id=('js', 'traverser', 'dangerous_global'),
                    warning='Access to the `%s` global' % name,
                    description='Access to the `%s` property is '
                                'deprecated for security or '
                                'other reasons.' % name)

                if isinstance(dang, DESCRIPTION_TYPES):
                    kwargs['description'] = dang
                elif isinstance(dang, dict):
                    kwargs.update(dang)

                self._debug('DANGEROUS')
                self.warning(**kwargs)

        entity.setdefault('name', name)

        # Build out the wrapper object from the global definition.
        result = JSWrapper(is_global=True, traverser=self, lazy=True)
        result.value = entity.copy()
        result = actions._expand_globals(self, result)

        if 'context' in entity:
            result.context = entity['context']

        self._debug('BUILT_GLOBAL')

        return result

    def _declare_variable(self, name, value, type_='var'):
        context = None
        if type_ in ('var', 'const', ):
            # Same as `reversed(self.contexts[1:])`
            for cont in itertools.islice(reversed(self.contexts), None,
                                         len(self.contexts)):
                if cont.type_ == 'default':
                    context = cont
                    break
            else:
                context = self.contexts[0]
        elif type_ == 'let':
            context = self.contexts[-1]
        elif type_ == 'glob':
            # Look down through the lexical scope. If the variable being
            # assigned is present in one of those objects, use that as the
            # target context.
            for ctx in reversed(self.contexts[1:]):
                if ctx.has_var(name):
                    context = ctx
                    break

        if not context:
            context = self.contexts[0]

        context.set(name, value)
        return value

    def _err_kwargs(self, kwargs):
        err_kwargs = {
            'filename': self.filename,
            'line': self.line,
            'column': self.position,
            'context': self.context,
        }
        err_kwargs.update(kwargs)
        return err_kwargs

    def error(self, **kwargs):
        err_kwargs = self._err_kwargs(kwargs)
        return self.err.error(**err_kwargs)

    def warning(self, **kwargs):
        err_kwargs = self._err_kwargs(kwargs)
        return self.err.warning(**err_kwargs)

    def notice(self, **kwargs):
        err_kwargs = self._err_kwargs(kwargs)
        return self.err.notice(**err_kwargs)

    def system_error(self, **kwargs):
        err_kwargs = self._err_kwargs(kwargs)
        return self.err.system_error(**err_kwargs)



from call_definitions import open_in_chrome_context
from instanceproperties import _set_HTML_property
from validator.compat import (
    FX47_DEFINITION, FX48_DEFINITION, FX50_DEFINITION, FX51_DEFINITION,
    FX52_DEFINITION, FX53_DEFINITION, FX54_DEFINITION)
from validator.constants import BUGZILLA_BUG, MDN_DOC


ENTITIES = {}


def register_entity(name):
    """Allow an entity's modifier to be registered for use."""
    def wrap(func):
        ENTITIES[name] = func
        return func
    return wrap


def entity(name, result=None):
    assert name in ENTITIES

    def return_wrap(t):
        output = ENTITIES[name](traverser=t)
        if result is not None:
            return result
        elif output is not None:
            return output
        else:
            return {'value': {}}
    return {'value': return_wrap}


def deprecated_entity(name, version, message, bug, status='deprecated',
                      compat_type='error'):
    def wrap(traverser):
        traverser.err.warning(
            err_id=('js', 'entities', name),
            warning='`%s` has been %s.' % (name, status),
            description=(message,
                         'See %s for more information.' % BUGZILLA_BUG % bug),
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context,
            for_appversions=version,
            compatibility_type=compat_type,
            tier=5)
    register_entity(name)(wrap)

def register_changed_entities(version_definition, entities, version_string):
    for entity in entities:
        deprecated_entity(
            name=entity['name'],
            version=version_definition,
            message='The method or property `%s` has been `%s` in `%s`.'
                % (entity['name'], entity['status'], version_string),
            bug=entity['bug'],
            compat_type=entity['compat_type'])


DOC_WRITE_MSG = ('https://developer.mozilla.org/docs/XUL/School_tutorial/'
                 'DOM_Building_and_HTML_Insertion')

@register_entity('document.write')
def document_write(traverser):
    def on_write(wrapper, arguments, traverser):
        traverser.err.warning(
            err_id=('js', 'document.write', 'evil'),
            warning='Use of `document.write` strongly discouraged.',
            description=('`document.write` will fail in many circumstances ',
                         'when used in extensions, and has potentially severe '
                         'security repercussions when used improperly. '
                         'Therefore, it should not be used. See %s for more '
                         'information.' % DOC_WRITE_MSG),
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)
        if not arguments:
            return
        value = traverser._traverse_node(arguments[0])
        _set_HTML_property('document.write()', value, traverser)

    return {'return': on_write}


@register_entity('nsIDNSService.resolve')
def nsIDNSServiceResolve(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIDNSServiceResolve'),
        warning='`nsIDNSService.resolve()` should not be used.',
        description='The `nsIDNSService.resolve` method performs a '
                    'synchronous DNS lookup, which will freeze the UI. This '
                    'can result in severe performance issues. '
                    '`nsIDNSService.asyncResolve()` should be used instead.',
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


@register_entity('nsISound.play')
def nsISoundPlay(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsISound_play'),
        warning='`nsISound.play` should not be used.',
        description='The `nsISound.play` function is synchronous, and thus '
                    'freezes the interface while the sound is playing. It '
                    'should be avoided in favor of the HTML5 audio APIs.',
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


@register_entity('nsIWindowWatcher.openWindow')
def nsIWindowWatcher_openWindow(traverser):
    def on_open(wrapper, arguments, traverser):
        if not arguments:
            return
        uri = traverser._traverse_node(arguments[0])
        open_in_chrome_context(uri, 'nsIWindowWatcher.openWindow', traverser)

    return {'return': on_open}


@register_entity('nsITransferable.init')
def nsITransferable_init(traverser):
    def on_init(wrapper, arguments, traverser):
        if not arguments:
            return
        first_arg = traverser._traverse_node(arguments[0])
        if first_arg.get_literal_value():
            return
        traverser.err.warning(
            err_id=('js_entity_values', 'nsITransferable', 'init'),
            warning='`init` should not be called with a null first argument',
            description='Calling `nsITransferable.init()` with a null first '
                        'argument has the potential to leak data across '
                        'private browsing mode sessions. `null` is  '
                        'appropriate only when reading data or writing data '
                        'which is not associated with a particular window.',
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)

    return {'return': on_init}


@register_entity('NewTabURL.override')
def NewTabURL_override(traverser):
    def on_override(wrapper, arguments, traverser):
        # Import loop.
        from validator.testcases.javascript.predefinedentities import (
            CUSTOMIZATION_API_HELP)
        traverser.err.warning(
            err_id=('js_entity_values', 'NewTabURL', 'override'),
            warning='Extensions must not alter user preferences such as the '
                    'new tab URL without explicit user consent.',
            description='Extensions must not alter user preferences such as '
                        'the new tab URL without explicit user consent. Such '
                        'changes must also be reverted when the extension is '
                        'disabled or uninstalled.',
            signing_severity='high',
            signing_help='Add-ons which directly change these preferences must '
                         'undergo manual code review for at least one '
                         'submission. ' + CUSTOMIZATION_API_HELP,
        )
    return {'return': on_override}


@register_entity('nsIObserverService.addObserver')
def nsIObserverService_addObserver(traverser):
    def on_addObserver(wrapper, arguments, traverser):
        if not arguments:
            return

        first_arg = traverser._traverse_node(arguments[1]).get_literal_value()

        if first_arg == 'newtab-url-changed':
            traverser.err.warning(
                err_id=('js_entity_values', 'nsIObserverService', 'newtab_url_changed'),
                warning='Extensions must not use the `newtab-url-changed` event to '
                        'revert changes made to the new tab url.',
                description='To avoid conflicts, extensions are not allowed to '
                            'add an observer to the `newtab-url-changed` event '
                            'in order to revert changes that have been made to '
                            'the new tab url.',
                signing_severity='high',
                signing_help='Add-ons which use `newtab-url-changed` to change '
                             'the new tab url are not allowed.')

    # Return the addObserver handler and a general dangerous warning.
    return {
        'return': on_addObserver,
        'dangerous': lambda a, t, e:
            e.get_resource('em:bootstrap') and
            'Authors of bootstrapped add-ons must take care '
            'to remove any added observers '
            'at shutdown.'
    }


@register_entity('nsIPK11TokenDB.listTokens')
@register_entity('nsIPKCS11ModuleDB.listModules')
@register_entity('nsIPKCS11Module.listSlots')
def nsIPK11TokenDB(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIPKThings'),
        warning='listTokens(), listModules() and listSlots() now return '
                'nsISimpleEnumerator instead of nsIEnumerator.',
        description=(
            'listTokens(), listModules() and listSlots() now return '
            'nsISimpleEnumerator instead of nsIEnumerator.'
            'See %s for more information.' % BUGZILLA_BUG % 1220237),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX47_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)


@register_entity('nsIIOService.newChannel')
@register_entity('nsIIOService.newChannelFromURI')
@register_entity('nsIIOService.newChannelFromURIWithProxyFlags')
def nsIIOService(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIIOService'),
        warning=(
            'The "newChannel" functions have been deprecated in favor of '
            'their new versions (ending with 2).'),
        description=(
            'The "newChannel" functions have been deprecated in favor of '
            'their new versions (ending with 2). '
            'See %s for more information.'
            % MDN_DOC % 'Mozilla/Tech/XPCOM/Reference/Interface/nsIIOService'),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX48_DEFINITION,
        compatibility_type='warning',
        context=traverser.context,
        tier=5)


@register_entity('nsIX509Cert.getUsagesArray')
@register_entity('nsIX509Cert.requestUsagesArrayAsync')
@register_entity('nsIX509Cert.getUsagesString')
def nsIX509Cert(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIX509Cert'),
        warning=(
            'The methods getUsagesArray, requestUsagesArrayAsync, and '
            'getUsagesString are no longer supported.'),
        description=(
            'The methods getUsagesArray, requestUsagesArrayAsync, and '
            'getUsagesString are no longer supported.',
            'See %s for more information.'
            % BUGZILLA_BUG % 1284946),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX50_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)


@register_entity('mozIAsyncFavicons.setAndFetchFaviconForPage')
@register_entity('mozIAsyncFavicons.replaceFaviconDataFromDataURL')
def mozIAsyncFavicons(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'mozIAsyncFavicons'),
        warning=(
            'The methods setAndFetchFaviconForPage and '
            'replaceFaviconDataFromDataURL now default to using a null '
            'principal for security reasons. An appropriate principal should '
            'be passed if different behavior is required.'),
        description=(
            'See %s for more information.'
            % BUGZILLA_BUG % 1227289),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX51_DEFINITION,
        compatibility_type='warning',
        context=traverser.context,
        tier=5)


@register_entity('nsISupportsArray')
def nsISupportsArray(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values', 'nsISupportsArray'),
        warning=(
            'The nsISupportsArray interface is deprecated and is being '
            'replaced by nsIArray.'),
        description=(
            'The nsISupportsArray interface is deprecated and is being '
            'replaced by nsIArray. See %s for more information.'
            % BUGZILLA_BUG % 792209),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX52_DEFINITION,
        compatibility_type='warning',
        context=traverser.context,
        tier=5)


@register_entity('nsINavBookmarksService.getURIForKeyword')
def nsINavBookmarksService(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsINavBookmarksService'),
        warning='The getURIForKeyword function was removed.',
        description=(
            'You can use PlacesUtils.keywords.fetch instead. '
            'See %s for more information.'
            % BUGZILLA_BUG % 1329926),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX53_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)


@register_entity('nsIX509CertDB.findCertByNickname')
@register_entity('nsIX509CertDB.findEmailEncryptionCert')
@register_entity('nsIX509CertDB.findEmailSigningCert')
@register_entity('nsIX509CertDB.addCert')
def nsIX509CertDB(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values',
                'nsIX509Cert'),
        warning=(
            'The nsIX509CertDB interface was changed so it no longer '
            'exposes the certificate nickname.'),
        description=(
            'The nsIX509CertDB interface was changed so it no longer '
            'exposes the certificate nickname. See %s for more information.'
            % BUGZILLA_BUG % 857627),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX53_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)


@register_entity('nsIPK11TokenDB.findTokenByName')
def nsIPK11TokenDBfindTokenByName(traverser):
    def _return(wrapper, arguments, traverser):
        if not arguments:
            return

        first_arg = traverser._traverse_node(arguments[0])
        link = 'https://hg.mozilla.org/mozilla-central/rev/6f732b38a5a2#l1.12'

        if first_arg.get_literal_value() == '':
            traverser.err.warning(
                err_id=('testcases_javascript_entity_values',
                        'nsIPK11TokenDB', 'findTokenByName'),
                warning='Calling findTokenByName("") is no longer valid.',
                description=(
                    'Calling findTokenByName("") is no longer valid. '
                    'If you need to determine if there\'s a master password '
                    'set, please see %s for more information.' % link),
                filename=traverser.filename,
                line=traverser.line,
                column=traverser.position,
                for_appversions=FX53_DEFINITION,
                compatibility_type='error',
                context=traverser.context,
                tier=5)
    return {'return': _return}


@register_entity(u'nsIFormHistory2')
def nsIFormHistory2(traverser):
    traverser.err.warning(
        err_id=('testcases_javascript_entity_values', 'nsIFormHistory2'),
        warning=(
            'The nsIFormHistory2 interface has been removed. You can use '
            'FormHistory.jsm instead.'),
        description=(
            'The nsIFormHistory2 interface has been removed. You can use '
            'FormHistory.jsm instead. See %s for more information.'
            % BUGZILLA_BUG % 876002),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        for_appversions=FX54_DEFINITION,
        compatibility_type='error',
        context=traverser.context,
        tier=5)



from helper import CompatTestCase
from validator.compat import FX48_DEFINITION


class TestFX48Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 48 are properly executed."""

    VERSION = FX48_DEFINITION

    def test_nsIIOService_newChannel_deprecated(self):
        """https://github.com/mozilla/addons-server/issues/2679"""
        expected = (
            'The "newChannel" functions have been deprecated in favor of '
            'their new versions (ending with 2).')

        script = '''
            var iOService = Components.classes["@mozilla.org/network/io-service;1"]
                .getService(Components.interfaces.nsIIOService);
            let channel = iOService.newChannel();
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'warning'

    def test_nsIIOService_newChannelFromURI_deprecated(self):
        """https://github.com/mozilla/addons-server/issues/2679"""
        expected = (
            'The "newChannel" functions have been deprecated in favor of '
            'their new versions (ending with 2).')

        script = '''
            var io = Components.classes["@mozilla.org/network/io-service;1"]
                .getService(Components.interfaces["nsIIOService"]);
            reqObj = io.newChannelFromURI(uri);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'warning'

    def test_nsIIOService_newChannelFromURIWithProxyFlags(self):
        """https://github.com/mozilla/addons-server/issues/2679"""
        expected = (
            'The "newChannel" functions have been deprecated in favor of '
            'their new versions (ending with 2).')

        script = '''
            var iOService = Components.classes["@mozilla.org/network/io-service;1"]
                .getService(Components.interfaces.nsIIOService);
            var channel = iOService.newChannelFromURIWithProxyFlags(aboutURI);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'warning'

    def test_Proxy_createFunction_removed(self):
        """https://github.com/mozilla/addons-server/issues/2678"""
        expected = 'Proxy.create and Proxy.createFunction are no longer supported.'

        script = '''
             function f() { "use strict"; return this; }

             var p = Proxy.createFunction(f, f);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_Proxy_create_removed(self):
        """https://github.com/mozilla/addons-server/issues/2678"""
        expected = 'Proxy.create and Proxy.createFunction are no longer supported.'

        script = '''
            var handler = {};

            var proxiedFunctionPrototype = Proxy.create(handler, Function.prototype, undefined);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        warning = [msg for msg in self.compat_err.warnings if msg['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_throbber_icon_moved(self):
        """https://github.com/mozilla/addons-server/issues/2678"""
        expected = 'The throbber icon your add-on points to has been moved'

        paths = (
            'chrome://browser/skin/tabbrowser/loading.png',
            'chrome://global/skin/icons/loading_16.png'
        )

        script = '''
            #foo {
                list-style-image: url("%s");
            }
        '''

        for path in paths:
            self.run_regex_for_compat(script % path, is_js=False)

            assert not self.compat_err.notices
            assert not self.compat_err.errors

            warning = [
                msg for msg in self.compat_err.warnings
                if msg['message'].startswith(expected)]
            assert warning
            assert warning[0]['compatibility_type'] == 'warning'



import actions

#(branches,
# explicitly_dynamic,
# estab_context,
# action,
# returns, # as in yielding a value, not breaking execution
# block_statement,
#)

def node(branches=None, dynamic=False, action=None, returns=False,
         is_block=False):
    if branches is None:
        branches = ()
    return branches, dynamic, action, returns, is_block


DEFINITIONS = {
    'EmptyStatement': node(),
    'DebuggerStatement': node(),

    'Program': node(branches=('body', ), is_block=True),
    'BlockStatement': node(branches=('body', ), is_block=True),
    'ExpressionStatement': node(branches=('expression', ),
                                action=actions._expression,
                                returns=True),
    'IfStatement': node(branches=('test', 'alternate', 'consequent'),
                        is_block=True),
    'LabeledStatement': node(branches=('body', )),
    'BreakStatement': node(),
    'ContinueStatement': node(),
    'WithStatement': node(branches=('body', 'object'),
                          action=actions._define_with, is_block=True),
    'SwitchStatement': node(branches=('test', 'cases'), is_block=True),
    'ReturnStatement': node(branches=('argument', )),
    'ThrowStatement': node(branches=('argument', )),
    'TryStatement': node(branches=('block', 'handler', 'finalizer',
                                   'guardedHandlers'),
                         is_block=True),
    'WhileStatement': node(branches=('test', 'body'), is_block=True),
    'DoWhileStatement': node(branches=('test', 'body'), is_block=True),
    'ForStatement': node(branches=('init', 'test', 'update', 'body'),
                         is_block=True),
    'ForInStatement': node(branches=('left', 'right', 'body'), is_block=True),
    'ForOfStatement': node(branches=('left', 'right', 'body'), is_block=True),

    'FunctionDeclaration': node(branches=('body', ), dynamic=True,
                                action=actions._define_function,
                                is_block=True),
    'VariableDeclaration': node(branches=('declarations', ),
                                action=actions._define_var),
    'LetStatement': node(branches=('head', 'body'),
                         action=actions._define_var),

    'ThisExpression': node(action=actions._get_this, returns=True),
    'ArrayExpression': node(branches=('elements', ),
                            action=actions._define_array, returns=True),
    'ObjectExpression': node(branches=('properties', ),
                             action=actions._define_obj, returns=True),
    'CallSiteObject': node(branches=('elements', ),
                           action=actions._define_template_strings,
                           returns=True),
    'TemplateLiteral': node(branches=('elements', ),
                            action=actions._define_template, returns=True),
    'FunctionExpression': node(branches=('body', ), dynamic=True,
                               action=actions._func_expr, returns=True,
                               is_block=True),
    'ArrowFunctionExpression': node(branches=('body', ), dynamic=True,
                                    action=actions._func_expr, returns=True,
                                    is_block=True),
    'SequenceExpression': node(branches=('expressions', ), returns=True),
    'UnaryExpression': node(branches=('argument', ),
                            action=actions._expr_unary, returns=True),
    'BinaryExpression': node(branches=('left', 'right'),
                             action=actions._expr_binary, returns=True),
    'AssignmentExpression': node(branches=('left', 'right'),
                                 action=actions._expr_assignment,
                                 returns=True),
    'UpdateExpression': node(branches=('argument', ), returns=True),
    'LogicalExpression': node(branches=('left', 'right'), returns=True),
    'ConditionalExpression': node(branches=('test', 'alternate', 'consequent'),
                                  returns=True),
    'NewExpression': node(branches=('constructor', 'arguments'),
                          action=actions._new, returns=True),
    'CallExpression': node(branches=('callee', 'arguments'),
                           action=actions._call_expression, returns=True),
    'TaggedTemplate': node(branches=('callee', 'arguments'),
                           action=actions._call_expression, returns=True),
    'MemberExpression': node(branches=('object', 'property'),
                             action=actions.trace_member, returns=True),
    'YieldExpression': node(branches=('argument',), returns=True),
    'ComprehensionExpression': node(branches=('body', 'filter'), returns=True),
    'GeneratorExpression': node(branches=('body', 'filter'), returns=True),

    'ObjectPattern': node(),
    'ArrayPattern': node(),

    'SwitchCase': node(branches=('test', 'consequent')),
    'CatchClause': node(branches=('param', 'guard', 'body'), returns=True),
    'ComprehensionBlock': node(branches=('left', 'right'), returns=True),

    'Literal': node(action=actions._define_literal, returns=True),
    'Identifier': node(action=actions._ident, returns=True),
    'GraphExpression': node(),
    'GraphIndexExpression': node(),
    'UnaryOperator': node(returns=True),
    'BinaryOperator': node(returns=True),
    'LogicalOperator': node(returns=True),
    'AssignmentOperator': node(returns=True),
    'UpdateOperator': node(returns=True),
}



from setuptools import setup, find_packages

from validator import __version__

setup(
    name='amo-validator',
    version=__version__,  # noqa
    description='Validates add-ons for Mozilla products.',
    long_description=open('README.md').read(),
    author='Matt Basta',
    author_email='me@mattbasta.com',
    url='http://github.com/mozilla/amo-validator',
    license='BSD',
    packages=find_packages(exclude=['tests', 'tests/*',
                                    'extras', 'extras/*']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[p.strip() for p in open('./requirements.txt') if
                      not p.startswith(('#', '-e'))],
    scripts=["addon-validator"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



import pytest

from helper import CompatTestCase
from validator.compat import FX52_DEFINITION


class TestFX52Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 52 are properly executed."""

    VERSION = FX52_DEFINITION

    def test_nsISupportsArray_deprecated(self):
        """https://github.com/mozilla/amo-validator/issues/493"""
        expected = (
            'The nsISupportsArray interface is deprecated and is being '
            'replaced by nsIArray.')

        script = '''
        var messages = Components.classes["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_warning('warning', expected)

    def test_mozDash_error(self):
        """https://github.com/mozilla/amo-validator/issues/492"""
        expected = 'The mozDash and mozDashOffset properties are no longer supported.'

        script = '''
            var canvas = document.getElementById('mycanvas');
            var ctx = canvas.getContext('2d');
            console.log(ctx.mozDash);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)

    def test_mozDashOffset_error(self):
        """https://github.com/mozilla/amo-validator/issues/492"""
        expected = 'The mozDash and mozDashOffset properties are no longer supported.'

        script = '''
            var canvas = document.getElementById('mycanvas');
            var ctx = canvas.getContext('2d');
            console.log(ctx.mozDashOffset);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        self.assert_compat_error('warning', expected)



import re
try:
    import curses
except ImportError:
    curses = None
import os
import sys

from StringIO import StringIO

COLORS = ('BLUE', 'RED', 'GREEN', 'YELLOW', 'WHITE', 'BLACK')


class OutputHandler:
    """A handler that hooks up with the error bundler to colorize the
    output of the application for *nix-based terminals."""

    def __init__(self, buffer=sys.stdout, no_color=False):
        if not curses:
            no_color = True
        if not no_color:
            no_color = isinstance(sys.stdout, StringIO) or \
                       not sys.stdout.isatty()

        self.no_color = no_color

        # Don't bother initiating color if there's no color.
        if not no_color:

            # Get curses all ready to write some stuff to the screen.
            curses.setupterm()

            # Initialize a store for the colors and pre-populate it
            # with the un-color color.
            self.colors = {'NORMAL': curses.tigetstr('sgr0') or ''}

            # Determines capabilities of the terminal.
            fgColorSeq = curses.tigetstr('setaf') or \
                curses.tigetstr('setf') or ''

            # Go through each color and figure out what the sequences
            # are for each, then store the sequences in the store we
            # made above.
            for color in COLORS:
                colorIndex = getattr(curses, 'COLOR_%s' % color)
                self.colors[color] = curses.tparm(fgColorSeq,
                                                  colorIndex)

        self.buffer = buffer

    def colorize_text(self, text):
        """Adds escape sequences to colorize text and make it
        beautiful. To colorize text, prefix the text you want to color
        with the color (capitalized) wrapped in double angle brackets
        (i.e.: <<GREEN>>). End your string with <<NORMAL>>. If you
        don't, it will be done for you (assuming you used a color code
        in your string."""

        # Take note of where the escape sequences are.
        rnormal = text.rfind('<<NORMAL')
        rany = text.rfind('<<')

        # Put in the escape sequences.
        for color, code in self.colors.items():
            text = text.replace('<<%s>>' % color, code)

        # Make sure that the last sequence is a NORMAL sequence.
        if rany > -1 and rnormal < rany:
            text += self.colors['NORMAL']

        return text

    def write(self, text):
        'Uses curses to print in the fanciest way possible.'

        # Add color to the terminal.
        if not self.no_color:
            text = self.colorize_text(text)
        else:
            pattern = re.compile('\<\<[A-Z]*?\>\>')
            text = pattern.sub('', text)

        text += '\n'

        self.buffer.write(text)

        return self

